INTEGER = 'INTEGER'  # 整数类型
REAL = 'REAL'  # 实数类型
INTEGER_CONST = 'INTEGER_CONST'  # 整数（因子）
REAL_CONST = 'REAL_CONST'  # 实数（因子）
PLUS = 'PLUS'  # 加
MINUS = 'MINUS'  # 减
MUL = 'MUL'  # 乘
INTEGER_DIV = 'INTEGER_DIV'  # 整数除法
FLOAT_DIV = 'FLOAT_DIV'  # 浮点数除法
LPAREN = 'LPAREN'  # 左括号
RPAREN = 'RPAREN'  # 右括号
ID = 'ID'  # 变量名称
ASSIGN = 'ASSIGN'  # 赋值符号
BEGIN = 'BEGIN'  # 开始标记
END = 'END'  # 结束标记
SEMI = 'SEMI'  # 分号
DOT = 'DOT'  # 点（程序结束符）
PROGRAM = 'PROGRAM'  # 程序
VAR = 'VAR'  # 变量声明标记
COLON = 'COLON'  # 冒号
COMMA = 'COMMA'  # 逗号
EOF = 'EOF'  # 结束符号


class Token:
    def __init__(self, value_type, value):
        self.value_type = value_type
        self.value = value

    def __str__(self):
        return 'Token({value_type},{value})'.format(value_type=self.value_type, value=self.value)

    def __repr__(self):
        return self.__str__()


RESERVED_KEYWORDS = {  # 保留字
    'PROGRAM': Token('PROGRAM', 'PROGRAM'),
    'VAR': Token('VAR', 'VAR'),
    'DIV': Token('INTEGER_DIV', 'DIV'),
    'INTEGER': Token('INTEGER', 'INTEGER'),
    'REAL': Token('REAL', 'REAL'),
    'BEGIN': Token('BEGIN', 'BEGIN'),
    'END': Token('END', 'END'),
}


class Lexer:
    def __init__(self, text):
        self.text = text
        self.position = 0
        self.current_char = self.text[self.position]

    def error(self):
        raise Exception('警告：错误的输入内容！')

    def number(self):  # 添加处理整数和实数的方法
        result = ''  # 定义保存数字字符串的变量
        while self.current_char is not None and self.current_char.isdigit():  # 如果当前字符不是None并且为整数
            result += self.current_char  # 连接已有数字字符串
            self.advance()  # 提取下一字符
            if self.current_char == '.':  # 如果当前字符为小数点
                result += self.current_char  # 连接已有数字字符串
                self.advance()  # 提取下一字符
                while self.current_char is not None and self.current_char.isdigit():  # 如果当前字符不是None并且为整数（小数点后均为整数）
                    result += self.current_char  # 连接已有数字字符串
                    self.advance()  # 提取下一字符
                return Token(REAL_CONST, float(result))  # 遍历完小数点右侧所有整数后，返回存储了小数数值的实数记号。
        return Token(INTEGER_CONST, int(result))  # 没有小数点时，遍历完所有整数后，返回存储了整数数值的整数记号。

    def skip_comment(self):  # 添加跳过注释内容到的方法
        while self.current_char != '}':  # 如果当前字符不是注释结束符号
            self.advance()  # 提取下一个字符
        self.advance()  # 提取下一个字符（跳过注释结束符号）

    def _id(self):
        result = ''
        if self.current_char == '_':  # 此语句块实现变量名下划线开头
            result = '_'
            self.advance()
        while self.current_char is not None and self.current_char.isalnum():
            result += self.current_char
            self.advance()
        token = RESERVED_KEYWORDS.get(result.upper(), Token('ID', result.lower()))  # 此语句块实现忽略关键字和变量名大小写
        return token

    def peek(self):
        pos = self.position + 1
        if pos >= len(self.text):
            return None
        else:
            return self.text[pos]

    def skip_whitespace(self):
        while self.current_char is not None and self.current_char.isspace():
            self.advance()

    # def long_integer(self): # 丢弃的长整数方法
    #     result = ''
    #     while self.current_char is not None and self.current_char.isdigit():
    #         result += self.current_char
    #         self.advance()
    #     return int(result)

    def advance(self):
        self.position += 1
        if self.position >= len(self.text):
            self.current_char = None
        else:
            self.current_char = self.text[self.position]

    def get_next_token(self):
        while self.current_char is not None:
            if self.current_char == '{':  # 如果当前字符为注释起始标记
                self.advance()  # 跳过当前字符
                self.skip_comment()  # 跳过注释内容
                continue  # 继续获取下一记号
            if self.current_char.isspace():
                self.skip_whitespace()
                continue
            if self.current_char.isalpha() or self.current_char == '_':  # 如果当前字符是字母或下划线（支持变量名称下划线开头）
                return self._id()
            if self.current_char == ':' and self.peek() == '=':
                self.advance()
                self.advance()
                return Token(ASSIGN, ':=')
            if self.current_char == ';':
                self.advance()
                return Token(SEMI, ';')
            if self.current_char == '.':
                self.advance()
                return Token(DOT, '.')
            if self.current_char.isdigit():  # 当前字符为整数时
                return self.number()  # 获取整数或实数记号
            if self.current_char == ':':  # 当前字符为冒号时
                self.advance()  # 提取下一字符
                return Token(COLON, ':')  # 返回冒号记号
            if self.current_char == ',':  # 当前字符为逗号时
                self.advance()  # 提取下一字符
                return Token(COMMA, ',')  # 返回逗号标记
            if self.current_char == '+':
                self.advance()
                return Token(PLUS, '+')
            if self.current_char == '-':
                self.advance()
                return Token(MINUS, '-')
            if self.current_char == '*':
                self.advance()
                return Token(MUL, '*')
            if self.current_char == '/':  # 当前字符为斜杠时
                self.advance()  # 提取下一字符
                return Token(FLOAT_DIV, '/')  # 返回浮点数除法记号
            if self.current_char == '(':
                self.advance()
                return Token(LPAREN, '(')
            if self.current_char == ')':
                self.advance()
                return Token(RPAREN, ')')
            self.error()
        return Token(EOF, None)


class AST:
    pass


class Program(AST):  # 定义程序节点
    def __init__(self, name, block):  # 程序由名称和语句块组成
        self.name = name
        self.block = block


class Block(AST):  # 定义语句块节点
    def __init__(self, declarations, compound_statement):  # 语句块由声明和符合语句组成
        self.declarations = declarations
        self.compound_statement = compound_statement


class VarDecl(AST):  # 定义变量声明节点
    def __init__(self, var_node, type_node):  # 变量声明由变量和类型组成
        self.var_node = var_node
        self.type_node = type_node


class Type(AST):  # 定义类型节点
    def __init__(self, token):
        self.token = token
        self.name = token.value


class Compound(AST):
    def __init__(self):
        self.children = []


class Assign(AST):
    def __init__(self, left, operator, right):
        self.left = left
        self.token = self.operator = operator
        self.right = right


class Variable(AST):
    def __init__(self, token):
        self.token = token
        self.name = token.value


class NoOperator(AST):
    pass


class BinaryOperator(AST):
    def __init__(self, left, operator, right):
        self.left = left
        self.token = self.operator = operator
        self.right = right


class Num(AST):
    def __init__(self, token):
        self.token = token
        self.value = token.value


class UnaryOperator(AST):
    def __init__(self, operator, expr):
        self.token = self.operator = operator
        self.expr = expr


class Parser:
    def __init__(self, lexer):
        self.lexer = lexer
        self.current_token = lexer.get_next_token()

    def error(self):
        raise Exception('警告：错误的输入内容！')

    def eat(self, value_type):
        if self.current_token.value_type == value_type:
            self.current_token = self.lexer.get_next_token()
        else:
            self.error()

    def program(self):
        self.eat(PROGRAM)  # 验证程序开始标记
        var_node = self.variable()  # 获取变量节点
        program_name = var_node.name  # 获取程序名称
        self.eat(SEMI)  # 验证分号
        block_node = self.block()  # 获取块节点
        node = Program(program_name, block_node)  # 创建程序节点
        self.eat(DOT)  # 验证程序结束符号
        return node  # 返回程序节点

    def block(self):  # 构造块节点的方法
        declarations = self.declarations()
        compound_statement = self.compound_statement()
        node = Block(declarations, compound_statement)  # 块节点由声明节点和符合语句节点组成
        return node

    def declarations(self):  # 构造声明节点的方法
        declarations = []  # 声明节点包含多个变量声明节点
        if self.current_token.value_type == VAR:  # 如果当前记号为变量
            self.eat(VAR)  # 验证记号
            while self.current_token.value_type == ID:  # 遍历变量名称
                declarations.extend(self.variable_declaration())  # 声明列表中添加变量声明
                self.eat(SEMI)  # 验证分号
        return declarations  # 返回声明节点列表

    def variable_declaration(self):  # 构造变量声明节点的方法
        var_nodes = [Variable(self.current_token)]  # 第一个变量声明节点添加到变量声明节点列表
        self.eat(ID)  # 验证变量名称记号
        while self.current_token.value_type == COMMA:  # 遍历逗号
            self.eat(COMMA)  # 验证逗号
            var_nodes.append(Variable(self.current_token))  # 添加变量节点到变量节点列表
            self.eat(ID)  # 验证变量名称记号
        self.eat(COLON)  # 验证冒号
        type_node = self.type_spec()  # 一组变量声明的类型节点
        var_declarations = [VarDecl(var_node, type_node) for var_node in var_nodes]  # 生成变量声明列表
        return var_declarations  # 返回变量声明节点列表

    def type_spec(self):  # 构造变量类型节点的方法
        token = self.current_token  # 获取当前记号
        if token.value_type == INTEGER:  # 如果是整数类型
            self.eat(INTEGER)  # 验证整数记号
        else:  # 否则
            self.eat(REAL)  # 验证实数记号
        node = Type(token)  # 创建类型节点
        return node  # 返回类型节点

    def compound_statement(self):
        self.eat(BEGIN)
        nodes = self.statement_list()
        self.eat(END)

        root = Compound()
        root.children = nodes
        return root

    def statement_list(self):
        node = self.statement()
        nodes = [node]
        while self.current_token.value_type == SEMI:
            self.eat(SEMI)
            nodes.append(self.statement())
        if self.current_token.value_type == ID:
            self.error()
        return nodes

    def statement(self):
        if self.current_token.value_type == BEGIN:
            node = self.compound_statement()
        elif self.current_token.value_type == ID:
            node = self.assignment_statement()
        else:
            node = self.empty()
        return node

    def assignment_statement(self):
        left = self.variable()
        token = self.current_token
        self.eat(ASSIGN)
        right = self.expr()
        node = Assign(left, token, right)
        return node

    def empty(self):
        return NoOperator()

    def variable(self):
        node = Variable(self.current_token)
        self.eat(ID)
        return node

    def factor(self):
        current_token = self.current_token

        if current_token.value_type == PLUS:
            self.eat(PLUS)
            node = UnaryOperator(current_token, self.factor())
            return node
        elif current_token.value_type == MINUS:
            self.eat(MINUS)
            node = UnaryOperator(current_token, self.factor())
            return node
        elif current_token.value_type == INTEGER_CONST:  # 整数
            self.eat(INTEGER_CONST)
            return Num(current_token)
        elif current_token.value_type == REAL_CONST:  # 实数
            self.eat(REAL_CONST)
            return Num(current_token)
        elif current_token.value_type == LPAREN:
            self.eat(LPAREN)
            node = self.expr()
            self.eat(RPAREN)
            return node
        else:
            node = self.variable()
            return node

    def term(self):
        node = self.factor()
        while self.current_token.value_type in (MUL, INTEGER_DIV, FLOAT_DIV):  # 除法修改为整数除法和浮点数除法
            token = self.current_token
            if token.value_type == MUL:
                self.eat(MUL)
            elif token.value_type == INTEGER_DIV:  # 整数除法
                self.eat(INTEGER_DIV)
            elif token.value_type == FLOAT_DIV:  # 浮点数除法
                self.eat(FLOAT_DIV)
            node = BinaryOperator(node, token, self.factor())
        return node

    def expr(self):
        node = self.term()
        while self.current_token.value_type in (PLUS, MINUS):
            token = self.current_token
            if token.value_type == PLUS:
                self.eat(PLUS)
            elif token.value_type == MINUS:
                self.eat(MINUS)
            node = BinaryOperator(node, token, self.term())
        return node

    def parser(self):
        node = self.program()
        if self.current_token.value_type != EOF:
            self.error()
        return node


class NodeVisitor:
    def visit(self, node):
        method_name = 'visit_' + type(node).__name__
        visitor = getattr(self, method_name, self.generic_visit)
        return visitor(node)

    def generic_visit(self, node):
        raise Exception(f'未找到“visit_{type(node).__name__}()”方法！')


class Interpreter(NodeVisitor):
    GLOBAL_SCOPE = {}

    def __init__(self, parser):
        self.parser = parser

    def visit_Program(self, node):  # 添加访问程序的方法
        self.visit(node.block)  # 访问语句块

    def visit_Block(self, node):  # 添加访问语句块的方法
        for declaration in node.declarations:  # 遍历声明列表
            self.visit(declaration)  # 访问声明
        self.visit(node.compound_statement)  # 访问复合语句

    def visit_VarDecl(self, node):  # 添加访问变量声明的方法
        pass  # 无需处理

    def visit_Type(self, node):  # 添加访问类型的方法
        pass  # 无需处理

    def visit_Compound(self, node):
        for child in node.children:
            self.visit(child)

    def visit_Assign(self, node):
        var_name = node.left.name
        self.GLOBAL_SCOPE[var_name] = self.visit(node.right)

    def visit_Variable(self, node):
        var_name = node.name
        value = self.GLOBAL_SCOPE.get(var_name)
        if value is None:
            raise NameError(f'错误的标识符：{repr(var_name)}')
        else:
            return value

    def visit_NoOperator(self, node):
        pass

    def visit_BinaryOperator(self, node):
        if node.operator.value_type == PLUS:
            return self.visit(node.left) + self.visit(node.right)
        elif node.operator.value_type == MINUS:
            return self.visit(node.left) - self.visit(node.right)
        elif node.operator.value_type == MUL:
            return self.visit(node.left) * self.visit(node.right)
        elif node.operator.value_type == INTEGER_DIV:  # 如果是整数除法
            return self.visit(node.left) // self.visit(node.right)  # 返回整数除法运算结果
        elif node.operator.value_type == FLOAT_DIV:  # 如果是浮点数除法
            return self.visit(node.left) / self.visit(node.right)  # 返回浮点数除法运算结果

    def visit_Num(self, node):
        return node.value

    def visit_UnaryOperator(self, node):
        if node.operator.value_type == PLUS:
            return +self.visit(node.expr)
        if node.operator.value_type == MINUS:
            return -self.visit(node.expr)

    def interpret(self):
        tree = self.parser.parser()
        return self.visit(tree)


def main():
    text = '''
    PROGRAM Part10;
    VAR
       number     : INTEGER;
       a, b, c, x : INTEGER;
       y          : REAL;
    
    BEGIN {Part10}
       BEGIN
          number := 2;
          a := number;
          b := 10 * a + 10 * number DIV 6;
          c := a - - b
       END;
       x := 11;
       y := 20 / 7 + 3.14;
       { writeln('a = ', a); }
       { writeln('b = ', b); }
       { writeln('c = ', c); }
       { writeln('number = ', number); }
       { writeln('x = ', x); }
       { writeln('y = ', y); }
    END.  {Part10}
    '''
    lexer = Lexer(text)
    parser = Parser(lexer)
    interpreter = Interpreter(parser)
    interpreter.interpret()
    print(interpreter.GLOBAL_SCOPE)


if __name__ == '__main__':
    main()
